/*
 * Copyright 2000 Hans Reiser
 */

#include "fsck.h"



#define bh_desc(bh) ((struct reiserfs_journal_desc *)((bh)->b_data))
#define bh_commit(bh) ((struct reiserfs_journal_commit *)((bh)->b_data))





static int next_expected_desc (reiserfs_filsys_t fs, struct buffer_head * d_bh)
{
    int offset;
    struct reiserfs_journal_desc * desc;

    desc = (struct reiserfs_journal_desc *)d_bh->b_data;
    offset = d_bh->b_blocknr - get_sb_journal_1st_block (fs->s_ondisk_sb);
    return get_sb_journal_1st_block (fs->s_ondisk_sb) + 
	((offset + get_jdesc_len (desc) + 1 + 1) % get_sb_journal_size (fs->s_ondisk_sb));
}


static int is_valid_transaction (reiserfs_filsys_t fs, struct buffer_head * d_bh)
{
    struct buffer_head * c_bh;
    int offset;
    struct reiserfs_journal_desc * desc = (struct reiserfs_journal_desc *)d_bh->b_data;
    struct reiserfs_journal_commit *commit ;
    __u32 block, start_block;


    offset = d_bh->b_blocknr - get_sb_journal_1st_block (fs->s_ondisk_sb);
    
    /* ok, we have a journal description block, lets see if the transaction was valid */
    block = next_expected_desc (fs, d_bh) - 1;
    start_block  = d_bh->b_blocknr;
    while(!(c_bh = bread (fs->s_journal_dev, block, fs->s_blocksize))){
        if (++block == get_sb_journal_1st_block (fs->s_ondisk_sb) + get_sb_journal_size (fs->s_ondisk_sb))
            block = get_sb_journal_1st_block (fs->s_ondisk_sb);
        if (block == start_block)
            return 0;
    }

    commit = (struct reiserfs_journal_commit *)c_bh->b_data ;
    if (does_desc_match_commit (desc, commit)) {
	brelse (c_bh) ;
	return 0;
    }
    brelse (c_bh);
    return 1;
}


int reiserfs_replay_journal (reiserfs_filsys_t fs)
{
    struct buffer_head * d_bh, * c_bh, * jh_bh;
    struct reiserfs_journal_header * j_head;
    struct reiserfs_journal_desc * j_desc;
    struct reiserfs_journal_commit * j_commit;
    unsigned long latest_mount_id;
    unsigned long j_cur;
    unsigned long j_start;
    unsigned long j_size;
    unsigned long mount_id, trans_id;
    unsigned long t_first, t_last, t_count, t_flushed;
    unsigned long t_offset;
    int i;

    fsck_progress ("Analyzing journal..");

    j_start = get_sb_journal_1st_block (fs->s_ondisk_sb);
    j_cur = 0;
    j_size = get_sb_journal_size (fs->s_ondisk_sb);
    t_first = 0;
    t_last = 0;
    latest_mount_id = 0;

    /* look for the transactions with the most recent mount_id */
    for (j_cur = 0; j_cur < j_size; ) {
	d_bh = bread (fs->s_journal_dev, j_start + j_cur, fs->s_blocksize);
	if (d_bh && who_is_this (d_bh->b_data, d_bh->b_size) == THE_JDESC &&
	    is_valid_transaction (fs, d_bh)) {
	    j_desc = (struct reiserfs_journal_desc *)d_bh->b_data;

	    mount_id = get_jdesc_mount_id (j_desc);
	    trans_id = get_jdesc_trans_id (j_desc);

	    if (mount_id > latest_mount_id) {
		/* more recent mount_id found */
		latest_mount_id = mount_id;
		t_first = t_last = trans_id;
		t_offset = j_cur;
		t_count = 1;
	    } else if (mount_id == latest_mount_id) {
		t_count ++;
		if (trans_id > t_last)
		    t_last = trans_id;
		if (trans_id < t_first) {
		    t_first = trans_id;
		    t_offset = j_cur;
		}
	    }
	    j_cur += get_jdesc_len (j_desc) + 1;
	}
	j_cur ++;
	brelse (d_bh);
    }

    /* replay only if journal header looks resonable */
    jh_bh = bread (fs->s_journal_dev, j_start + j_size, fs->s_blocksize);
    j_head = (struct reiserfs_journal_header *)(jh_bh->b_data);

    if (latest_mount_id != le32_to_cpu (j_head->j_mount_id)) {
	fsck_progress ("nothing to replay (no transactions match to latest mount id)\n");
	brelse (jh_bh);
	return 0;
    }
    /* last transaction flushed - which should not be replayed */
    t_flushed = le32_to_cpu (j_head->j_last_flush_trans_id);
    if (t_flushed >= t_last) {
	fsck_progress ("nothing to replay (no transactions older than last flushed one found)\n");
	brelse (jh_bh);
	return 0;
    }
    if (t_first > t_flushed + 1) {
	if (t_flushed)
	    fsck_progress ("last flushed trans %lu, the oldest but newer is %lu\n",
			   t_flushed, t_first);
    } else {
	/* start replaying with first not flushed transaction */
	t_first = t_flushed + 1;
	t_offset = le32_to_cpu (j_head->j_first_unflushed_offset);
    }

    fsck_progress ("last flushed trans %lu, mount_id %lu, "
		   "will replay from %lu up to %lu:Yes?",
		   t_flushed, latest_mount_id, t_first, t_last);
    if (!fsck_user_confirmed (fs, "", "Yes\n", 1))
	die ("");

    /* replay transactions we have found */
    for (j_cur = t_offset; t_first <= t_last; t_first ++) {
	unsigned long offset;
	
	d_bh = bread (fs->s_journal_dev, j_start + j_cur, fs->s_blocksize);
	j_desc = (struct reiserfs_journal_desc *)d_bh->b_data;
	if (who_is_this (d_bh->b_data, d_bh->b_size) != THE_JDESC ||
	    get_jdesc_mount_id (j_desc) != latest_mount_id ||
	    get_jdesc_trans_id (j_desc) != t_first)
	    die ("reiserfs_replay_journal: desc block not found");

	offset = j_cur + 1;
	j_cur += get_jdesc_len (j_desc) + 1;
	j_cur %= j_size;
	c_bh = bread (fs->s_journal_dev, j_start + j_cur, fs->s_blocksize);
	j_commit = (struct reiserfs_journal_commit *)c_bh->b_data;
	if (does_desc_match_commit (j_desc, j_commit))
	    die ("reiserfs_replay_journal: commit block not found");

	fsck_log ("Mount_id %lu, transaction %lu, desc block %lu, commit block %lu: (",
		  latest_mount_id, t_first, d_bh->b_blocknr, c_bh->b_blocknr);

	/* replay one transaction */
	for (i = 0; i < get_jdesc_len (j_desc); i ++) {
	    struct buffer_head * in_place, * log;
	    unsigned long block;

	    log = bread (fs->s_journal_dev, j_start + ((offset + i) % j_size), fs->s_blocksize);

	    if (i < JOURNAL_TRANS_HALF) {
		block = le32_to_cpu (j_desc->j2_realblock[i]);
	    } else {
		block = le32_to_cpu (j_commit->j3_realblock[i - JOURNAL_TRANS_HALF]);
	    }

	    if (not_journalable (fs, block)) {
		fsck_log ("transaction %lu, block %d could not be replayed (%lu)\n",
			      t_first, i, block);
	    } else {
		fsck_log (" %lu", block);
		
		in_place = getblk (fs->s_dev, block, fs->s_blocksize) ;
		memcpy (in_place->b_data, log->b_data, fs->s_blocksize);
		mark_buffer_dirty (in_place);
		mark_buffer_uptodate (in_place, 1);
		bwrite (in_place);
		brelse (in_place);
	    }
	    brelse (log);
	    
	}
	fsck_log (")\n");

	brelse (d_bh);
	brelse (c_bh);
	j_cur ++;
	j_cur %= j_size;

	/* update journal header */
	j_head->j_last_flush_trans_id = cpu_to_le32 (t_first);
	mark_buffer_dirty (jh_bh);
	bwrite (jh_bh);
    }

    brelse (jh_bh);
    fsck_progress ("Journal replaied\n");
    return 0;
}
